#include <config.h>
#include <math.h>
#include <string.h>

#include <glib/gi18n.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#include <gdk-pixbuf/gdk-pixbuf.h>

#include "button-widget.h"


#define BUTTON_WIDGET_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), BUTTON_TYPE_WIDGET, ButtonWidgetPrivate))

struct _ButtonWidgetPrivate {
	GdkPixbuf        *pixbuf;
	GdkPixbuf        *pixbuf_hc;

	TalikaTask *task;

	guint             dnd_highlight : 1;
};

static void button_widget_reload_pixbuf (ButtonWidget *button);

enum {
	PROP_0,
	PROP_DND_HIGHLIGHT
};

#define BUTTON_WIDGET_DISPLACEMENT 2

G_DEFINE_TYPE (ButtonWidget, button_widget, GTK_TYPE_TOGGLE_BUTTON)

/* colorshift a pixbuf */
static void
do_colorshift (GdkPixbuf *dest, GdkPixbuf *src, int shift)
{
	gint i, j;
	gint width, height, has_alpha, srcrowstride, destrowstride;
	guchar *target_pixels;
	guchar *original_pixels;
	guchar *pixsrc;
	guchar *pixdest;
	int val;
	guchar r,g,b;

	has_alpha = gdk_pixbuf_get_has_alpha (src);
	width = gdk_pixbuf_get_width (src);
	height = gdk_pixbuf_get_height (src);
	srcrowstride = gdk_pixbuf_get_rowstride (src);
	destrowstride = gdk_pixbuf_get_rowstride (dest);
	target_pixels = gdk_pixbuf_get_pixels (dest);
	original_pixels = gdk_pixbuf_get_pixels (src);

	for (i = 0; i < height; i++) {
		pixdest = target_pixels + i*destrowstride;
		pixsrc = original_pixels + i*srcrowstride;
		for (j = 0; j < width; j++) {
			r = *(pixsrc++);
			g = *(pixsrc++);
			b = *(pixsrc++);
			val = r + shift;
			*(pixdest++) = CLAMP(val, 0, 255);
			val = g + shift;
			*(pixdest++) = CLAMP(val, 0, 255);
			val = b + shift;
			*(pixdest++) = CLAMP(val, 0, 255);
			if (has_alpha)
				*(pixdest++) = *(pixsrc++);
		}
	}
}

static GdkPixbuf *
make_hc_pixbuf (GdkPixbuf *pb)
{
	GdkPixbuf *new;
	
	if (!pb)
		return NULL;

	new = gdk_pixbuf_new (gdk_pixbuf_get_colorspace (pb),
			      gdk_pixbuf_get_has_alpha (pb),
			      gdk_pixbuf_get_bits_per_sample (pb),
			      gdk_pixbuf_get_width (pb),
			      gdk_pixbuf_get_height (pb));
	do_colorshift (new, pb, 30);

	return new;
}

static void
button_widget_realize(GtkWidget *widget)
{
	GdkWindowAttr attributes;
	gint attributes_mask;
	GtkButton *button;

	g_return_if_fail (widget != NULL);
	g_return_if_fail (BUTTON_IS_WIDGET (widget));

	button = GTK_BUTTON (widget);

	GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);

	attributes.window_type = GDK_WINDOW_CHILD;
	attributes.x = widget->allocation.x;
	attributes.y = widget->allocation.y;
	attributes.width = widget->allocation.width;
	attributes.height = widget->allocation.height;
	attributes.wclass = GDK_INPUT_ONLY;
	attributes.event_mask = (GDK_BUTTON_PRESS_MASK |
				 GDK_BUTTON_RELEASE_MASK |
				 GDK_POINTER_MOTION_MASK |
				 GDK_POINTER_MOTION_HINT_MASK |
				 GDK_KEY_PRESS_MASK |
				 GDK_ENTER_NOTIFY_MASK |
				 GDK_LEAVE_NOTIFY_MASK);
	attributes_mask = GDK_WA_X | GDK_WA_Y;

	widget->window = gtk_widget_get_parent_window (widget);
	g_object_ref (G_OBJECT (widget->window));
      
	button->event_window = gdk_window_new (gtk_widget_get_parent_window (widget),
					       &attributes,
					       attributes_mask);
	gdk_window_set_user_data (button->event_window, widget);

	widget->style = gtk_style_attach (widget->style, widget->window);

	button_widget_reload_pixbuf (BUTTON_WIDGET (widget));
}

static void
button_widget_unrealize (GtkWidget *widget)
{
	GtkButton *button;

	g_return_if_fail (widget != NULL);
	g_return_if_fail (BUTTON_IS_WIDGET (widget));

	button = GTK_BUTTON (widget);

	if (button->event_window != NULL) {
		gdk_window_set_user_data (button->event_window, NULL);
		gdk_window_destroy (button->event_window);
		button->event_window = NULL;
	}

	GTK_WIDGET_CLASS (button_widget_parent_class)->unrealize (widget);
}

static void
button_widget_unset_pixbufs (ButtonWidget *button)
{
	if (button->priv->pixbuf)
		g_object_unref (button->priv->pixbuf);
	button->priv->pixbuf = NULL;

	if (button->priv->pixbuf_hc)
		g_object_unref (button->priv->pixbuf_hc);
	button->priv->pixbuf_hc = NULL;
}

static void
button_widget_reload_pixbuf (ButtonWidget *button)
{
	//button_widget_unset_pixbufs (button);
	
	if (button->priv->pixbuf_hc)
		g_object_unref (button->priv->pixbuf_hc);
	button->priv->pixbuf_hc = NULL;

	button->priv->pixbuf_hc = make_hc_pixbuf (button->priv->pixbuf);

	gtk_widget_queue_resize (GTK_WIDGET (button));
}

static void
button_widget_finalize (GObject *object)
{
	ButtonWidget *button = (ButtonWidget *) object;

	button_widget_unset_pixbufs (button);

	G_OBJECT_CLASS (button_widget_parent_class)->finalize (object);
}

static void
button_widget_dispose (GObject *object)
{
	ButtonWidget *button = (ButtonWidget *) object;

	button_widget_unset_pixbufs (button);

	G_OBJECT_CLASS (button_widget_parent_class)->dispose (object);
}

static void
button_widget_get_property (GObject    *object,
			    guint       prop_id,
			    GValue     *value,
			    GParamSpec *pspec)
{
	ButtonWidget *button;

	g_return_if_fail (BUTTON_IS_WIDGET (object));

	button = BUTTON_WIDGET (object);

	switch (prop_id) {
	case PROP_DND_HIGHLIGHT:
		g_value_set_boolean (value, button->priv->dnd_highlight);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
button_widget_set_property (GObject      *object,
			    guint         prop_id,
			    const GValue *value,
			    GParamSpec   *pspec)
{
	ButtonWidget *button;

	g_return_if_fail (BUTTON_IS_WIDGET (object));

	button = BUTTON_WIDGET (object);

	switch (prop_id) {
	case PROP_DND_HIGHLIGHT:
		button_widget_set_dnd_highlight (button, g_value_get_boolean (value));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void 
draw_indicator(cairo_t *cr, double x, double y,
                      double radius, GdkColor colorbg)
{
	cairo_move_to (cr,x,y);
	cairo_arc (cr, x, y, radius, 0, 2 * M_PI);

	double r = (double)colorbg.red / 65535;
	double g = (double)colorbg.green / 65535;
	double b = (double)colorbg.blue / 65535;
	
	//radial pattren
	cairo_pattern_t *rg = cairo_pattern_create_radial (x,y,0,x,y,radius);
	//cairo_pattern_add_color_stop_rgba (rg,0,1,1,1,1);
	cairo_pattern_add_color_stop_rgba (rg,0.20,r,g,b,1);
	cairo_pattern_add_color_stop_rgba (rg,0.30,r,g,b,0.60);
	cairo_pattern_add_color_stop_rgba (rg,0.60,r,g,b,0.25);
	cairo_pattern_add_color_stop_rgba (rg,1.0,r,g,b,0);
	

	cairo_set_source (cr,rg);
	//cairo_set_source_rgb (cr, 1, 1, 1);
	cairo_fill_preserve (cr);
	//cairo_set_source_rgb (cr, 0, 0, 0);
	//cairo_stroke (cr);
	cairo_pattern_destroy (rg);
}

static void
get_indicator_xy (PanelAppletOrient orient, gint width, gint height, 
                  double xfactor, double yfactor, 
                  double *x, double *y,double *radius)
{
	switch (orient)
	{
		case PANEL_APPLET_ORIENT_UP:
		{
			*radius = MIN (width / 2, height / 10);
			*x = *x + width * xfactor;
			*y = *y + height - height * yfactor;
		}
			break;

		case PANEL_APPLET_ORIENT_DOWN:
		{
			*radius = MIN (width / 2, height / 5);
			*x = *x + width * xfactor;
			*y = *y + height * yfactor;
		}
			break;

		case PANEL_APPLET_ORIENT_RIGHT:
		{
			*radius = MIN (width / 10, height / 2);
			*x = *x + width * yfactor;
			*y = *y + height * xfactor;
		}
			break;

		case PANEL_APPLET_ORIENT_LEFT:
		{
			*radius = MIN (width / 10, height / 2);
			*x = *x + width - width * yfactor;
			*y = *y + height * xfactor;
		}
			break;

		default:
			break;
	}
}

static void 
draw_activity_indicator (GtkWidget *widget, cairo_t *cr, GdkColor colorbg)
{
	double x,y,radius;

	//GtkStyle *style = gtk_widget_get_style (widget);
	//GdkColor colorbg = style->bg[GTK_STATE_SELECTED];

	ButtonWidget *button = BUTTON_WIDGET(widget);
	if(button->priv->task->type == TALIKA_TASK_WINDOW)
	{
		//x = widget->allocation.x + widget->allocation.width / 2;
		//y = widget->allocation.y + widget->allocation.height / 20;
		x = widget->allocation.x;
		y = widget->allocation.y;
		get_indicator_xy (button->priv->task->tasklist->panel_orient,
		                  widget->allocation.width,widget->allocation.height,
		                  0.5,0.05,
		                  &x,&y,&radius);
		draw_indicator(cr,x,y,radius,colorbg);
	}
	else
	{
		//class group
		x = widget->allocation.x;
		y = widget->allocation.y;
		get_indicator_xy (button->priv->task->tasklist->panel_orient,
		                  widget->allocation.width,widget->allocation.height,
		                  0.4,0.05,
		                  &x,&y,&radius);
		draw_indicator(cr,x,y,radius,colorbg);

		x = widget->allocation.x;
		y = widget->allocation.y;
		get_indicator_xy (button->priv->task->tasklist->panel_orient,
		                  widget->allocation.width,widget->allocation.height,
		                  0.6,0.05,
		                  &x,&y,&radius);
		draw_indicator(cr,x,y,radius,colorbg);
	}
}

static gboolean
button_widget_expose (GtkWidget         *widget,
		      GdkEventExpose    *event)
{
	ButtonWidget *button_widget;
	GtkButton *button;
	GtkToggleButton *toggle_button;
	GdkRectangle area, image_bound;
	int off;
	int x, y, w, h;
	GdkPixbuf *pb = NULL;
  
	g_return_val_if_fail (BUTTON_IS_WIDGET (widget), FALSE);
	g_return_val_if_fail (event != NULL, FALSE);

	button_widget = BUTTON_WIDGET (widget);
	button = GTK_BUTTON (widget);
	toggle_button = GTK_TOGGLE_BUTTON (widget);
	
	if (!GTK_WIDGET_VISIBLE (widget) || !GTK_WIDGET_MAPPED (widget))
		return FALSE;

	if (!button_widget->priv->pixbuf_hc && !button_widget->priv->pixbuf)
		return FALSE;

	if (toggle_button->active) {
		gint focus_width, focus_pad;
		gint x, y, width, height;

		gtk_widget_style_get (widget,
				      "focus-line-width", &focus_width,
				      "focus-padding", &focus_pad,
				      NULL);
		x = widget->allocation.x + focus_pad;
		y = widget->allocation.y + focus_pad;
		width = widget->allocation.width -  2 * focus_pad;
		height = widget->allocation.height - 2 * focus_pad;
		gtk_paint_focus (widget->style, widget->window,
				 GTK_STATE_NORMAL,
				 &event->area, widget, "button",
				 x, y, width, height);
	}
	
	//offset for pressed buttons 
	off = (button->in_button && button->button_down) ?
		BUTTON_WIDGET_DISPLACEMENT * widget->allocation.height / 48.0 : 0;

	if ((button->in_button))
		pb = g_object_ref (button_widget->priv->pixbuf_hc);
	else
		pb = g_object_ref (button_widget->priv->pixbuf);

	g_assert (pb != NULL);

	w = gdk_pixbuf_get_width (pb);
	h = gdk_pixbuf_get_height (pb);
	x = widget->allocation.x + off + (widget->allocation.width - w)/2;
	y = widget->allocation.y + off + (widget->allocation.height - h)/2;
	
	image_bound.x = x;
	image_bound.y = y;      
	image_bound.width = w;
	image_bound.height = h;
	
	area = event->area;
	
	if (gdk_rectangle_intersect (&area, &widget->allocation, &area) &&
	    gdk_rectangle_intersect (&image_bound, &area, &image_bound))
		gdk_draw_pixbuf (widget->window, NULL, pb,
				 image_bound.x - x, image_bound.y - y,
				 image_bound.x, image_bound.y,
				 image_bound.width, image_bound.height,
				 GDK_RGB_DITHER_NORMAL,
				 0, 0);

	g_object_unref (pb);
	
	

	if (button_widget->priv->dnd_highlight) {
		gdk_draw_rectangle(widget->window, widget->style->black_gc, FALSE,
				   widget->allocation.x, widget->allocation.y,
				   widget->allocation.width - 1,
				   widget->allocation.height - 1);
	}

	//draw activity indicators 
	if(button_widget->priv->task->tasklist->running_indicator &&
	   button_widget->priv->task->type != TALIKA_TASK_LAUNCHER)
	{
		cairo_t *cr;

		// get a cairo_t 
		cr = gdk_cairo_create (widget->window);

		// set a clip region for the expose event 
		cairo_rectangle (cr,
		                 widget->allocation.x, widget->allocation.y,
		                 widget->allocation.width, widget->allocation.height);
		cairo_clip (cr);

		draw_activity_indicator (widget, cr, 
		           button_widget->priv->task->tasklist->running_indicator_color);

		cairo_destroy (cr);
	}
	
	return FALSE;
}

static void
button_widget_size_request (GtkWidget      *widget,
			    GtkRequisition *requisition)
{
	ButtonWidget *button_widget = BUTTON_WIDGET (widget);

	if (button_widget->priv->pixbuf) {
		requisition->width  = button_widget->priv->task->tasklist->panel_size;
		requisition->height = button_widget->priv->task->tasklist->panel_size;

		//requisition->width  = gdk_pixbuf_get_width  (button_widget->priv->pixbuf);
		//requisition->height = gdk_pixbuf_get_height (button_widget->priv->pixbuf);
	}
}

static void
button_widget_size_allocate (GtkWidget     *widget,
			     GtkAllocation *allocation)
{
	//ButtonWidget *button_widget = BUTTON_WIDGET (widget);
	GtkButton    *button = GTK_BUTTON (widget);
	/*int           size;

	if (button_widget->priv->orientation & PANEL_HORIZONTAL_MASK)
		size = allocation->height;
	else
		size = allocation->width;

	if (size < 22)
		size = 16;
	else if (size < 24)
		size = 22;
	else if (size < 32)
		size = 24;
	else if (size < 48)
		size = 32;
	else
		size = 48;

	if (button_widget->priv->size != size) {
		button_widget->priv->size = size;

		button_widget_reload_pixbuf (button_widget);
	}*/
	
	//g_print ("%d,%d\n",allocation->width,allocation->height);
	widget->allocation = *allocation;

	if (GTK_WIDGET_REALIZED (widget)) {
		gdk_window_move_resize (button->event_window, 
					allocation->x,
					allocation->y,
					allocation->width,
					allocation->height);
	}
}

static gboolean
button_widget_button_press (GtkWidget *widget, GdkEventButton *event)
{
	g_return_val_if_fail (BUTTON_IS_WIDGET (widget), FALSE);
	g_return_val_if_fail (event != NULL, FALSE);

	if (event->button == 1 &&
	/* we don't want to have two/three "click" events for double/triple
	 * clicks. FIXME: this is only a workaround, waiting for bug 159101 */
	    event->type == GDK_BUTTON_PRESS)
		return GTK_WIDGET_CLASS (button_widget_parent_class)->button_press_event (widget, event);

	return FALSE;
}

static gboolean
button_widget_enter_notify (GtkWidget *widget, GdkEventCrossing *event)
{
	gboolean in_button;

	g_return_val_if_fail (BUTTON_IS_WIDGET (widget), FALSE);

	in_button = GTK_BUTTON (widget)->in_button;
	GTK_WIDGET_CLASS (button_widget_parent_class)->enter_notify_event (widget, event);
	if (in_button != GTK_BUTTON (widget)->in_button)
		gtk_widget_queue_draw (widget);

	return FALSE;
}

static gboolean
button_widget_leave_notify (GtkWidget *widget, GdkEventCrossing *event)
{
	gboolean in_button;

	g_return_val_if_fail (BUTTON_IS_WIDGET (widget), FALSE);

	in_button = GTK_BUTTON (widget)->in_button;
	GTK_WIDGET_CLASS (button_widget_parent_class)->leave_notify_event (widget, event);
	if (in_button != GTK_BUTTON (widget)->in_button)
		gtk_widget_queue_draw (widget);

	return FALSE;
}

static void
button_widget_init (ButtonWidget *button)
{
	button->priv = BUTTON_WIDGET_GET_PRIVATE (button);

	button->priv->pixbuf     = NULL;
	button->priv->pixbuf_hc  = NULL;

	button->priv->task = NULL;
	
	button->priv->dnd_highlight = FALSE;
}

static void
button_widget_class_init (ButtonWidgetClass *klass)
{
	GObjectClass *gobject_class   = (GObjectClass   *) klass;
	GtkWidgetClass *widget_class  = (GtkWidgetClass *) klass;
	//GtkToggleButtonClass *button_class  = (GtkToggleButtonClass *) klass;

	gobject_class->finalize     = button_widget_finalize;
	gobject_class->dispose     = button_widget_dispose;
	gobject_class->get_property = button_widget_get_property;
	gobject_class->set_property = button_widget_set_property;

	g_type_class_add_private (klass, sizeof (ButtonWidgetPrivate));
	  
	widget_class->realize            = button_widget_realize;
	widget_class->unrealize          = button_widget_unrealize;
	widget_class->size_allocate      = button_widget_size_allocate;
	widget_class->size_request       = button_widget_size_request;
	widget_class->button_press_event = button_widget_button_press;
	widget_class->enter_notify_event = button_widget_enter_notify;
	widget_class->leave_notify_event = button_widget_leave_notify;
	widget_class->expose_event       = button_widget_expose;

	g_object_class_install_property (
			gobject_class,
			PROP_DND_HIGHLIGHT,
			g_param_spec_boolean ("dnd-highlight",
					      "Drag and drop Highlight",
					      "Whether or not to highlight the icon during drag and drop",
					      FALSE,
					      G_PARAM_READWRITE));

}

GtkWidget *
button_widget_new (GdkPixbuf   *pixbuf,TalikaTask *task)
{
	GtkWidget *retval;

	retval = g_object_new (
			BUTTON_TYPE_WIDGET,
			NULL);
	
	button_widget_set_pixbuf (BUTTON_WIDGET(retval),pixbuf);

	BUTTON_WIDGET(retval)->priv->task = task;
	
	return retval;
}

void
button_widget_set_pixbuf (ButtonWidget *button,
			     GdkPixbuf   *pixbuf)
{
	g_return_if_fail (BUTTON_IS_WIDGET (button));

	if (!pixbuf)
		return;

	button_widget_unset_pixbufs (button);

	button->priv->pixbuf = g_object_ref(pixbuf);
	
	button_widget_reload_pixbuf (button);
}

void
button_widget_set_dnd_highlight (ButtonWidget *button,
				 gboolean      dnd_highlight)
{
	g_return_if_fail (BUTTON_IS_WIDGET (button));

	dnd_highlight = dnd_highlight != FALSE;

	if (button->priv->dnd_highlight == dnd_highlight)
		return;

	button->priv->dnd_highlight = dnd_highlight;

	gtk_widget_queue_draw (GTK_WIDGET (button));

	g_object_notify (G_OBJECT (button), "dnd-highlight");
}

gboolean
button_widget_get_dnd_highlight (ButtonWidget *button)
{
	g_return_val_if_fail (BUTTON_IS_WIDGET (button), FALSE);

	return button->priv->dnd_highlight;
}

GdkPixbuf *
button_widget_get_pixbuf (ButtonWidget *button)
{
	g_return_val_if_fail (BUTTON_IS_WIDGET (button), NULL);

	if (!button->priv->pixbuf)
		return NULL;

	return g_object_ref (button->priv->pixbuf);
}
